package node

import (
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"os"
	"os/exec"
	"sync"
	"text/template"
	"time"

	"git.autistici.org/ale/autoradio"
)

var (
	liquidsoapBin = flag.String("liquidsoap", "/usr/bin/liquidsoap", "Location of the liquidsoap binary")

	liquidsoapConfigStr = `
set("log.file", false)
set("log.stdout", true)
set("log.unix_timestamps", false)

upstream = mksafe(input.http("{{.SourceURL}}", buffer=5.0))
output.icecast(%{{.Format}}(samplerate={{.SampleRate}}, {{if gt .BitRate 0}}bitrate={{.BitRate}}, {{end}}{{if gt .Quality -1.0}}quality={{.Quality}}, {{end}}{{if .StereoMode}}stereo_mode="{{.StereoMode}}", {{end}}{{if eq .Channels 2}}stereo{{else}}mono{{end}}),
  mount="{{.TargetMount}}", host="{{.TargetIP}}", port={{.TargetPort}}, user="{{.TargetUsername}}", password="{{.TargetPassword}}",
  upstream)
`

	liquidsoapConfigTpl = template.Must(template.New("liquidsoap").Parse(liquidsoapConfigStr))
)

// Parameters to configure a liquidsoap-based transcoder.
type liquidsoapParams struct {
	// Source (upstream) URL.
	SourceURL string

	// Target (downstream) connection parameters.
	TargetIP       string
	TargetPort     int
	TargetMount    string
	TargetUsername string
	TargetPassword string

	// Stream encoding parameters. One note on the 'float32'
	// choice for Quality: text/template can't evaluate the 'gt'
	// function successfully on a float64 type!
	Format     string
	BitRate    int
	SampleRate int
	Channels   int
	StereoMode string
	Quality    float32
}

// Create new parameters for liquidsoap for a transcoding mount. If
// mount.Transcoding is nil, this function will panic, so the caller
// should check mount.HasTranscoder().
//
// The source URL points at localhost, relying on the front-end
// proxying to actually send the traffic to the master. This has the
// advantage that we don't have to reconfigure liquidsoap on
// mastership changes.
func newLiquidsoapParams(mount *autoradio.Mount) *liquidsoapParams {
	return &liquidsoapParams{
		SourceURL:      fmt.Sprintf("http://localhost%s", mount.Transcoding.SourceName),
		TargetIP:       "localhost",
		TargetPort:     80,
		TargetMount:    mount.Name,
		TargetUsername: mount.Username,
		TargetPassword: mount.Password,
		Format:         mount.Transcoding.Format,
		BitRate:        mount.Transcoding.BitRate,
		SampleRate:     mount.Transcoding.SampleRate,
		Channels:       mount.Transcoding.Channels,
		StereoMode:     mount.Transcoding.StereoMode,
		Quality:        float32(mount.Transcoding.Quality),
	}
}

func (p *liquidsoapParams) Equal(other *liquidsoapParams) bool {
	return *p == *other
}

func (p *liquidsoapParams) Render(w io.Writer) error {
	return liquidsoapConfigTpl.Execute(w, p)
}

// Controls a single liquidsoap process, transcoding a single stream.
// This object will be created once, when the node finds out about a
// transcoded stream for the first time. From then on, the Start() and
// Stop() methods will be called, possibly multiple times, whenever
// the node acquires or loses mastership for the stream. Specifically,
// this means that it must be possible to call Start() again after a
// Stop().
type liquidsoapController struct {
	params       *liquidsoapParams
	restartDelay time.Duration

	// Information about the running process. The lock protects
	// the 'process' field, and it is only unlocked when we're
	// waiting for the process to exit, or if it is not running at
	// all.
	stop    chan struct{}
	wg      sync.WaitGroup
	lock    sync.Mutex
	process *os.Process
}

func newLiquidsoap(params *liquidsoapParams) (*liquidsoapController, error) {
	return &liquidsoapController{
		params:       params,
		restartDelay: 500 * time.Millisecond,
	}, nil
}

// Creates a new temporary file with the Liquidsoap transcoding
// configuration, and returns its name.
func (l *liquidsoapController) writeConfig() (string, error) {
	file, err := ioutil.TempFile("", "liquidsoap-")
	if err != nil {
		return "", err
	}
	defer file.Close()
	return file.Name(), l.params.Render(file)
}

func (l *liquidsoapController) runOnce() {
	config, err := l.writeConfig()
	// Always remove the temporary file, if it was created, even
	// if there was an error.
	if config != "" {
		defer os.Remove(config)
	}
	if err != nil {
		log.Printf("error writing liquidsoap configuration file: %v", err)
		return
	}

	// Keep restarting liquidsoap if it fails, until the stop
	// channel is closed (requesting termination). Keep the lock
	// held until l.process is set, so that we can synchronize the
	// process termination properly.
	cmd := exec.Command(*liquidsoapBin, "-T", "-U", "-v", config)
	cmd.Stdout = os.Stderr
	cmd.Stderr = os.Stderr
	if err := cmd.Start(); err != nil {
		log.Printf("error starting liquidsoap: %v", err)
		return
	}
	l.process = cmd.Process

	// Release the lock while the process is running.
	l.lock.Unlock()
	err = cmd.Wait()
	l.lock.Lock()

	l.process = nil
	if err != nil {
		log.Printf("liquidsoap exited: %v", err)
	}
}

func (l *liquidsoapController) run() {
	defer l.wg.Done()

	// Called with l.lock being held, release it when done.
	defer l.lock.Unlock()

	for {
		// Limit the restart rate in case of errors.
		select {
		case <-l.stop:
			return
		case <-time.After(l.restartDelay):
		}

		l.runOnce()
	}
}

func (l *liquidsoapController) Start() {
	l.lock.Lock()
	l.wg.Add(1)
	l.stop = make(chan struct{})
	go l.run()
}

func (l *liquidsoapController) Stop() {
	// Close the stop channel to prevent restarts.
	close(l.stop)

	// Kill the currently running process, if any.
	l.lock.Lock()
	if l.process != nil {
		l.process.Kill()
	}
	l.lock.Unlock()

	l.wg.Wait()
}
